# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
from AthenaTwitchLib.irc.logic._logic import BaseCommandLogic, register_callback_as_logical_component
from AthenaTwitchLib.irc.logic.commands import CommandLogic, CommandData
from AthenaTwitchLib.irc.logic.commands_sqlite import CommandLogicSqlite
from AthenaTwitchLib.irc.logic.tasks import TaskLogic, TaskData


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass
from typing import Literal, ClassVar

# Athena Packages

# Local Imports
from AthenaTwitchLib.irc.tags._tags import Conversion, Tags, TAG_TYPES

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True, frozen=True)
class TagsUSERSTATE(Tags):
    """
    Class for Twitch IRC Tags, that are from the USERSTATE message
    """
    badge_info:str=None
    badges:str=None
    color:str=None
    display_name:str=None
    emote_sets: str = None
    id:str=None
    mod:bool=None
    subscriber:bool=None
    turbo:bool=None
    user_type:Literal["", "admin", "global_mod", "staff"]=None

    _tag_type:ClassVar[TAG_TYPES] = TAG_TYPES.USERSTATE
    _CONVERSION_MAPPING:ClassVar[dict] = {
        "badge-info": Conversion("badge_info",str),
        "badges": Conversion("badges",lambda obj: obj.split(",")),
        "bits": Conversion("bits",str),
        "color": Conversion("color",str),
        "display-name": Conversion("display_name",str),
        "emote-sets": Conversion("emote_sets",str),
        "id": Conversion("id",str),
        "mod": Conversion("mod",lambda obj: bool(int(obj))),
        "subscriber":Conversion("subscriber", lambda obj: bool(int(obj))),
        "turbo": Conversion("turbo",lambda obj: bool(int(obj))),
        "user-type": Conversion("user_type",str),
    }


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass, asdict
import asyncio

# Athena Packages

# Local Imports
from AthenaTwitchLib.irc.tags import TagsPRIVMSG
from AthenaTwitchLib.string_formatting import twitch_irc_output_format

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True, kw_only=True, frozen=True)
class MessageContext:
    """
    Frozen Dataclass which holds the context that will be used by the LogicBot to handle an incoming message
    """
    tags:TagsPRIVMSG
    user:str
    username:str
    channel:str
    text:str

    transport:asyncio.Transport
    bot_event_future:asyncio.Future
    original_line:str

    def as_dict(self) -> dict:
        """
        Casts the object to a dict that is usable in a JSON format
        """
        return {
            "tags": asdict(self.tags),
            "user": self.user,
            "username": self.username,
            "channel": self.channel,
            "text": self.text
        }

    async def reply(self, reply_msg:str):
        """
        Replies to the given message of the channel where it came from:
        """
        self.transport.write(
            twitch_irc_output_format(f"@reply-parent-msg-id={self.tags.id} PRIVMSG #{self.channel} :{reply_msg}")
        )

    async def write(self, write_msg:str):
        """
        Writes a message to the channel it came from:
        """
        self.transport.write(
            twitch_irc_output_format(f"PRIVMSG #{self.channel} :{write_msg}")
        )

# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True,kw_only=True, frozen=True)
class MessageCommandContext(MessageContext):
    """
    Frozen Dataclass which holds the context that will be used by the LogicBot to handle an incoming message
    The message in question should be a command
    """
    command:str
    args:list[str]

    def __post_init__(self):
        if self.args == ['']:
            self.args.clear()



# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass
from typing import ClassVar
import datetime

# Athena Packages

# Local Imports
from AthenaTwitchLib.irc.tags._tags import Conversion, Tags, TAG_TYPES

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True, frozen=True)
class TagsCLEARCHAT(Tags):
    """
    Class for Twitch IRC Tags, that are from the CLEARCHAT message
    """
    ban_duration:int=None
    room_id:str=None
    target_user_id:str=None
    tmi_sent_ts: datetime.datetime=None

    _tag_type:ClassVar[TAG_TYPES] = TAG_TYPES.CLEARCHAT
    _CONVERSION_MAPPING:ClassVar[dict] = {
        "ban-duration": Conversion("ban_duration",int),
        "room-id": Conversion("room_id",int),
        "target-user-id": Conversion("target_user_id",str),
        "tmi-sent-ts": Conversion("tmi_sent_ts",datetime.datetime.fromtimestamp),
    }


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass, field, asdict
from typing import Callable
import json

# Athena Packages

# Local Imports
from AthenaTwitchLib.irc.logic._logic import BaseHardCodedLogic, BaseCommandLogic,register_callback_as_logical_component
from AthenaTwitchLib.irc.message_context import MessageCommandContext
from AthenaTwitchLib.irc.tags import TagsPRIVMSG
from AthenaTwitchLib.logger import SectionIRC, IrcLogger

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True)
class CommandData:
    """
    Simple dataclass to hold basic information about the hard coded command.
    Meant for the `CommandLogic` class to differentiate what to do when it receives a command from chat
    ===

    - cmd_names:             All possible command Names
    - allow_user:            Boolean option to allow all users to use this command
    - allow_sub:             Boolean option to allow subscribers to use this command
    - allow_vip:             Boolean option to allow VIP's to use this command
    - allow_mod:             Boolean option to allow moderators to use this command
    - allow_broadcaster:     Boolean option to allow broadcasters to use this command

    """
    cmd_names:list[str]|str
    allow_user:bool=field(kw_only=True, default=True)
    allow_sub:bool=field(kw_only=True, default=False)
    allow_vip:bool=field(kw_only=True, default=False)
    allow_mod:bool=field(kw_only=True, default=False)
    allow_broadcaster:bool=field(kw_only=True, default=False)

    def __post_init__(self):
        if isinstance(self.cmd_names, str):
            self.cmd_names = [self.cmd_names]
        elif not isinstance(self.cmd_names, list):
            raise ValueError

        # Log to db
        IrcLogger.log_debug(
            section=SectionIRC.CMD_DATA,
            text=json.dumps(asdict(self))
        )

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
class CommandLogic(BaseHardCodedLogic,BaseCommandLogic):
    """
    Logic system behind hard coded commands.
    """
    _commands: dict[str: Callable]

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls, *args, *kwargs)
        obj._commands = {
            name:fnc
            for fnc in obj._logic_components
            for name in fnc._data.cmd_names
        }

        return obj

    async def execute_command(self, context:MessageCommandContext):
        """
        Main entry point from the Async Protocol, will try and find a corresponding command within the Bot.
        """
        # Get the command from the stored bot's method.
        #   If it can't be found, skip the entire function
        if not (fnc := self._commands.get(context.command, False)):
            IrcLogger.log_debug(
                section=SectionIRC.CMD_UNKNOWN,
                text=context.original_line
            )
            return

        # When a callback is found
        #   Execute command
        match fnc._data, context:

            # a command that all users can access
            case CommandData(allow_user=True), _:
                await fnc(context)

            case CommandData(allow_broadcaster=True), MessageCommandContext(user=user, channel=channel) if user == f":{channel}!{channel}@{channel}.tmi.twitch.tv":
                await fnc(context)

            case CommandData(allow_mod=True), MessageCommandContext(tags=TagsPRIVMSG(moderator=True)):
                await fnc(context)

            case CommandData(allow_sub=True), MessageCommandContext(tags=TagsPRIVMSG(subscriber=True)):
                await fnc(context)

            case CommandData(allow_vip=True), MessageCommandContext(tags=TagsPRIVMSG(vip=True)):
                await fnc(context)

            # in any other cases
            #   This should never happen
            case _,_:
                IrcLogger.log_warning(
                    section=SectionIRC.CMD_NOT_PARSABLE,
                    text=context.original_line
                )

    @staticmethod
    def command(command_data:CommandData):
        """
        Decorator to be used by the Bot, to assign a method as a command.
        """
        def decorator(fnc):
            register_callback_as_logical_component(fnc)
            fnc._data = command_data

            return fnc
        return decorator

    @staticmethod
    def command_subscriber(command_data:CommandData):
        """
        Decorator to be used by the Bot, to assign a method as a command only to be used by a Subscriber.
        The command will also be able to be used by Mods and Broadcasters
        """
        command_data.allow_user = False
        command_data.allow_sub = True
        command_data.allow_vip = False
        command_data.allow_mod = True
        command_data.allow_broadcaster = True

        def decorator(fnc):
            register_callback_as_logical_component(fnc)
            fnc._data = command_data

            return fnc
        return decorator

    @staticmethod
    def command_vip(command_data:CommandData):
        """
        Decorator to be used by the Bot, to assign a method as a command only to be used by a VIP.
        The command will also be able to be used by Mods and Broadcasters
        """
        command_data.allow_user = False
        command_data.allow_sub = False
        command_data.allow_vip = True
        command_data.allow_mod = True
        command_data.allow_broadcaster = True

        def decorator(fnc):
            register_callback_as_logical_component(fnc)
            fnc._data = command_data

            return fnc
        return decorator

    @staticmethod
    def command_moderator(command_data:CommandData):
        """
        Decorator to be used by the Bot, to assign a method as a command only to be used by a Moderator.
        The command will also be able to be used by Broadcasters
        """
        command_data.allow_user = False
        command_data.allow_sub = False
        command_data.allow_vip = True
        command_data.allow_mod = True
        command_data.allow_broadcaster = True

        def decorator(fnc):
            register_callback_as_logical_component(fnc)
            fnc._data = command_data

            return fnc
        return decorator

    @staticmethod
    def command_broadcaster(command_data:CommandData):
        """
        Decorator to be used by the Bot, to assign a method as a command only to be used by a Broadcasters.
        """
        command_data.allow_user = False
        command_data.allow_sub = False
        command_data.allow_vip = False
        command_data.allow_mod = False
        command_data.allow_broadcaster = True

        def decorator(fnc):
            register_callback_as_logical_component(fnc)
            fnc._data = command_data

            return fnc
        return decorator


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations

# Athena Packages

# Local Imports

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
TBL_LOGIC_COMMANDS = f"""
CREATE TABLE IF NOT EXISTS `commands`  (
    `id` INTEGER PRIMARY KEY,                               -- internal id
    `channel` TEXT NOT NULL,
    `command_name` TEXT NOT NULL,                           -- ![command_name]
    `command_arg` TEXT NOT NULL DEFAULT '*',                -- if this specific arg is present, execute this record, else everything is allowed
    `command_type` TEXT NOT NULL,                           -- Type of command, means it has extra logic attached to it
    `allow_user` INTEGER NOT NULL DEFAULT TRUE,
    `allow_sub` INTEGER NOT NULL DEFAULT FALSE,
    `allow_vip` INTEGER NOT NULL DEFAULT FALSE,
    `allow_mod` INTEGER NOT NULL DEFAULT FALSE,
    `allow_broadcaster` INTEGER NOT NULL DEFAULT FALSE,
    `output_text` TEXT DEFAULT NULL,                        -- text to output in chat
    `output_type` TEXT NOT NULL DEFAULT 'write'             -- reply or write output
);
"""


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass, field
import asyncio

# Athena Packages

# Local Imports
from AthenaTwitchLib.irc.logic import CommandLogic, TaskLogic
from AthenaTwitchLib.logger import SectionIRC, IrcLogger
from AthenaTwitchLib.string_formatting import twitch_irc_output_format

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True)
class Bot:
    """
    Base bot class for the connection
    """
    name:str
    oath_token:str
    join_channel:list[str] = field(default_factory=list)
    join_message:str = None
    prefix:str = "!"

    capability_tags:bool=True,
    capability_commands:bool=False,
    capability_membership:bool=False,

    # KW only
    command_logic:CommandLogic = field(kw_only=True, default_factory=CommandLogic)
    task_logic:TaskLogic = field(kw_only=True, default_factory=TaskLogic)

    # non init
    transport:asyncio.BaseTransport|asyncio.Transport = field(init=False)

    # ------------------------------------------------------------------------------------------------------------------
    # - Login functions -
    # ------------------------------------------------------------------------------------------------------------------
    async def _cap_tags(self):
        """
        Assigns the Twitch IRC chat capability of receiving tags
        This should always be requested, else answering to chat is impossible
        """
        if not self.capability_tags:
            IrcLogger.log_debug(section=SectionIRC.LOGIN_CAPABILITY, text="capability_tags not enabled")
            return

        self.transport.write(twitch_irc_output_format(f"CAP REQ :twitch.tv/tags"))
        IrcLogger.log_debug(section=SectionIRC.LOGIN_CAPABILITY, text="capability_tags set")

    async def _cap_commands(self):
        """
        Assigns the Twitch IRC chat capability of sending twitch (`/`) commands in chat, by the bot
        """
        if not self.capability_commands:
            IrcLogger.log_debug(section=SectionIRC.LOGIN_CAPABILITY, text="capability_commands not enabled")
            return

        self.transport.write(twitch_irc_output_format(f"CAP REQ :twitch.tv/commands"))
        IrcLogger.log_debug(section=SectionIRC.LOGIN_CAPABILITY, text="capability_commands set")

    async def _cap_membership(self):
        """
        Assigns the Twitch IRC chat capability of receiving membership information
        """
        if not self.capability_membership:
            IrcLogger.log_debug(section=SectionIRC.LOGIN_CAPABILITY, text="capability_membership not enabled")
            return

        self.transport.write(twitch_irc_output_format(f"CAP REQ :twitch.tv/membership"))
        IrcLogger.log_debug(section=SectionIRC.LOGIN_CAPABILITY, text="capability_membership set")

    async def _write_to_twitch(self, section:SectionIRC, txt:str):
        """
        Simple function that writes the text to the twitch chat
        """
        self.transport.write(twitch_irc_output_format(txt))
        IrcLogger.log_debug(section=section, text=txt)

    async def login(self):
        """
        Steps that need to be taken for the Bot to be logged into the Twitch IRC chat
        """
        # Login into the irc chat
        #   Not handled by the protocol,
        #   as it is a direct write only feature and doesn't need to respond to anything
        self.transport.write(twitch_irc_output_format(f"PASS oauth:{self.oath_token}"))
        self.transport.write(twitch_irc_output_format(f"NICK {self.name}"))

        # Log that we have logged in
        IrcLogger.log_debug(
            section=SectionIRC.LOGIN,
            text=f"[{self.name=}, {self.join_channel=}, {self.join_message=}, {self.prefix=}]"
        )

        # Request correct capabilities
        await asyncio.gather(
            # Join all channels and don't wait for the logger to finish
            *(self._write_to_twitch(section=SectionIRC.JOIN, txt=f"JOIN #{channel}")
              for channel in self.join_channel),

            # Get capabilities from Twitch
            self._cap_tags(),
            self._cap_commands(),
            self._cap_membership(),
        )

        # will catch all those that are Truthy (not: "", None, False, ...)
        if self.join_message:
            await asyncio.gather(
                (self._write_to_twitch(
                    section=SectionIRC.LOGIN_MSG,
                    txt=f"PRIVMSG #{channel} :{self.join_message}"
                )
                for channel in self.join_channel),
            )



# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
from AthenaTwitchLib.irc.tags._tags import Tags, TAG_TYPES
from AthenaTwitchLib.irc.tags.tags_clearchat import TagsCLEARCHAT
from AthenaTwitchLib.irc.tags.tags_clearmsg import TagsCLEARMSG
from AthenaTwitchLib.irc.tags.tags_globaluserstate import TagsGLOBALUSERSTATE
from AthenaTwitchLib.irc.tags.tags_notice import TagsNOTICE
from AthenaTwitchLib.irc.tags.tags_privmsg import TagsPRIVMSG
from AthenaTwitchLib.irc.tags.tags_roomstate import TagsROOMSTATE
from AthenaTwitchLib.irc.tags.tags_usernotice import TagsUSERNOTICE
from AthenaTwitchLib.irc.tags.tags_userstate import TagsUSERSTATE


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from dataclasses import dataclass
from typing import ClassVar, Literal

# Athena Packages

# Local Imports
from AthenaTwitchLib.irc.tags._tags import Conversion, Tags, TAG_TYPES

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
@dataclass(slots=True, frozen=True)
class TagsGLOBALUSERSTATE(Tags):
    """
    Class for Twitch IRC Tags, that are from the GLOBALUSERSTATE message
    """
    badge_info:str=None
    badges:str=None
    color:str=None
    display_name:str=None
    emote_set:str=None
    turbo:bool=None
    user_id:str=None
    user_type:Literal["", "admin", "global_mod", "staff"]=None

    _tag_type:ClassVar[TAG_TYPES] = TAG_TYPES.GLOBALUSERSTATE
    _CONVERSION_MAPPING:ClassVar[dict] = {
        "badge-info": Conversion("badge_info",str),
        "badges": Conversion("badges",lambda obj: obj.split(",")),
        "bits": Conversion("bits",str),
        "color": Conversion("color",str),
        "display-name": Conversion("display_name",str),
        "emote-set": Conversion("emote_set",str),
        "turbo": Conversion("turbo",lambda obj: bool(int(obj))),
        "user-id": Conversion("user_id",str),
        "user-type": Conversion("user_type",str)
    }